/*-----------------------------------------------------------------------
	Warp-GGI

	General Graphics Interface demo
	Realtime picture 'gooing'

	dowarp.c: warp routines

	written by Emmanuel Marty <emarty@mirus.fr>
	November 1st, 1997

	Released under GNU Public License
-----------------------------------------------------------------------*/

#include <sys/types.h>
#include <stdlib.h>
#include <math.h>
#include "warp.h"

void initSinTable (struct warp *w) {
	int	*tptr, *tsinptr;
	double	i;

	tsinptr = tptr = w->sintable;

	for (i = 0; i < 1024; i++)
		*tptr++ = (int) (sin (i*M_PI/512) * 32767);

	for (i = 0; i < 256; i++)
		*tptr++ = *tsinptr++;
}

void initOffsTable (struct warp *w) {
	int	width, height, len, y;
	char	*source;
	void	**offptr;

	offptr = w->offstable;
	width  = w->width; height = w->height;
	source = (char *) w->source;
	len    = w->srclinelen;

	for (y = 0; y < height; y++) {
		*offptr++ = (void *) source;
		source += len;
	}
}
      
void initDistTable (struct warp *w) {
	int	halfw, halfh, *distptr;
	double	x,y,m;

	halfw = w->width >> 1;
	halfh = w->height >> 1;

	distptr = w->disttable;

	m = sqrt ((double)(halfw*halfw + halfh*halfh));

	for (y = -halfh; y < halfh; y++)
		for (x= -halfw; x < halfw; x++)
			*distptr++ = ((int)
				( (sqrt (x*x+y*y) * 511.9999) / m)) << 1;
}

struct warp *initWarp (int width, int height, int depth,
		       void *source, int srclinelen) {
	int		bytespp;
	struct warp	*w;

	if ( (w = (struct warp *) malloc (sizeof (struct warp))) ) {
		if ( (w->offstable = malloc (height * sizeof (char *))) ) {
			bytespp = (depth == 8) ? 1 : 4;

			if ( (w->disttable = malloc (width *
                                                     height *
						  sizeof (int))) ) {
				if ( (w->framebuf = malloc (width *
							   height *
							   bytespp)) ) {
					w->width = width;
					w->height = height;
					w->depth = depth;
					w->source = source;
					w->srclinelen = srclinelen;

					initSinTable (w);
					initOffsTable (w);
					initDistTable (w);

					return (w);
				}
				free (w->disttable);
			}
			free (w->offstable);
		}
		free (w);
	}
	return (NULL);
}

void disposeWarp (struct warp *w) {
	if (w) {
		free (w->framebuf);
		free (w->disttable);
		free (w->offstable);
		free (w);
	}
}

void doWarp8bpp (struct warp *w, int xw, int yw, int cw) {
	int c,i, x,y, dx,dy, maxx, maxy;
	int width, height, *ctable, *ctptr, *distptr;
	int *sintable, *disttable;
	char *destptr, **offstable;

	ctptr = ctable = &(w->ctable[0]);
	sintable = &(w->sintable[0]);
	offstable = (char **) w->offstable;
	distptr = disttable = w->disttable;
	width = w->width;
	height = w->height;
	destptr = (char *) w->framebuf;

	c = 0;

	for (x = 0; x < 512; x++) {
		i = (c >> 3) & 0x3FE;
		*ctptr++ = ((sintable[i] * yw) >> 15);
		*ctptr++ = ((sintable[i+256] * xw) >> 15);
		c += cw;
	}

	maxx = width - 1; maxy = height - 1;

	for (y = 0; y < height; y++)
         for (x = 0; x < width; x++) {
		i = *distptr++;
		dx = ctable [i+1] + x;
		dy = ctable [i] + y;
                                
		if (dx < 0) dx = 0;
		else if (dx > maxx) dx = maxx;

		if (dy < 0) dy = 0;
		else if (dy > maxy) dy = maxy;
                                
                *destptr++ = * (offstable[dy] + dx);
        }
}
